# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for performing parsing of web pages."""

from __future__ import annotations

import asyncio
import collections
import dataclasses

import aiohttp
import bs4

from googleads_housekeeper.domain.external_parsers import base_parser


@dataclasses.dataclass
class WebsiteInfo(base_parser.EntityInfo):
  """Contains meta information of websites.

  Attributes:
      placement: URL of website's domain.
      title: Title of website.
      description: Website description from 'meta' tags.
      keywords: Website keywords from 'meta' tags.
      is_processed: Whether website was successfully parsed.
  """

  placement: str
  title: str = ''
  description: str = ''
  keywords: str = ''
  is_processed: bool = True


class WebSiteParser(base_parser.BaseParser):
  """Performs website parsing."""

  def parse(
    self, placements: collections.abc.Sequence[str]
  ) -> list[WebsiteInfo]:
    """Parses provided websites.

    Args:
        placements: Sequence of website domains.

    Returns:
        Parsed information in WebsiteInfo format.
    """
    return asyncio.run(self._parse_placements(placements))

  async def _parse_placements(
    self, placements: collections.abc.Sequence[str]
  ) -> list[WebsiteInfo]:
    """Parses provided websites asynchronously.

    Args:
        placements: Sequence of website domains.

    Returns:
        Parsed information in WebsiteInfo format.
    """
    timeout = aiohttp.ClientTimeout(total=10)
    async with aiohttp.ClientSession(timeout=timeout) as session:
      results = []
      for placement in placements:
        results.append(self._parse_placement(placement, session))
      values = await asyncio.gather(*results)
    return values

  async def _parse_placement(
    self, placement: str, session: aiohttp.ClientSession
  ) -> WebsiteInfo:
    """Parses single website.

    Args:
        placement: Website domain.

    Returns:
        Parsed information in WebsiteInfo format.
    """
    url = self._convert_placement_to_url(placement)
    try:
      html = await self._fetch_html(url, session)
      placement_info = await self._extract_from_html(placement, html)
      return placement_info
    except Exception:
      return WebsiteInfo(placement=placement, is_processed=False)

  def _convert_placement_to_url(self, placement: str) -> str:
    """Normalizes url."""
    if 'http' not in placement:
      url = f'https://{placement}'
    else:
      url = placement
    return url

  async def _fetch_html(self, url: str, session: aiohttp.ClientSession):
    """Gets website content."""
    response = await session.request(method='GET', url=url)
    response.raise_for_status()
    html = await response.text()
    return html

  async def _extract_from_html(self, url: str, html: str) -> WebsiteInfo:
    """Extracts tags from html."""
    soup = bs4.BeautifulSoup(html, 'html.parser')
    keywords = soup.find('meta', attrs={'name': 'keywords'})
    description = soup.find('meta', attrs={'name': 'description'})
    return WebsiteInfo(
      placement=url,
      title=soup.title.string if soup.title else None,
      keywords=keywords.get('content') if keywords else None,
      description=description.get('content') if description else None,
      is_processed=True,
    )



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for defining an interface for parsers."""

from __future__ import annotations

import abc
import collections
import dataclasses
import datetime
import uuid
from typing import TypeVar


@dataclasses.dataclass
class EntityInfo:
  """Contains base information on parsed entity.

  Attributes:
      placement: Name of parsed placement.
      last_processed_time: Current time when processing happened.
      id: Unique identifier of placement.
  """

  placement: str
  last_processed_time: datetime = dataclasses.field(
    default=datetime.datetime.now(), compare=False
  )
  id: str = dataclasses.field(
    default_factory=lambda: str(uuid.uuid4()), compare=False
  )


T = TypeVar('T', bound='EntityInfo')


class BaseParser(abc.ABC):
  """Base class for all external parsers."""

  @abc.abstractmethod
  def parse(self, placements: collections.abs.Sequence[str]) -> list[T]:
    """Parses provided placements.

    Placement could be website URL, YouTube channel or video id.

    Args:
        placements: Sequence of placements identifiers.

    Returns:
        List of parsed entities in subclass *Info format.
    """
    raise NotImplementedError


class NullParser(BaseParser):
  """Special case of invalid parser."""

  def parse(self, placements):
    return placements



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for defining preview_results objects."""

from __future__ import annotations

# pylint: disable=C0330
import dataclasses
import datetime
import enum
import uuid
from typing import Optional, Union


class PreviewTaskStatus(enum.Enum):
  """Holds type of Task execution."""

  RUNNING = "RUNNING"
  DONE = "DONE"
  ERROR = "ERROR"


@dataclasses.dataclass
class PreviewTaskAndResults:
  """Holds information of a particular preview result set.

  Attributes:
      status: execution status
      creation_time: Time when execution started.
      preview_command: The preview command
      results: Optional result details as a text field.
      id: Unique identifier of execution.
  """

  preview_command: str
  creation_time: datetime = dataclasses.field(
    default_factory=datetime.datetime.now
  )
  status: PreviewTaskStatus = PreviewTaskStatus.RUNNING
  results: Optional[str] = None
  id: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))

  DATETIME_FORMAT = "%d-%b-%Y %H:%M"


  def to_dict(self) -> dict[str, Union[str, None]]:
      """Convert the object to a Firestore-compatible dictionary."""
      return {
          "preview_command": self.preview_command,
          "creation_time": self.creation_time.strftime(self.DATETIME_FORMAT),
          "status": self.status,
          "results": self.results,
          "id": self.id,
      }

  @staticmethod
  def from_dict(data: dict[str, Union[str, None]]) -> (
          PreviewTaskAndResults):
      """Convert Firestore dictionary to a PreviewTaskAndResults object."""
      return PreviewTaskAndResults(
          preview_command=data["preview_command"],
          creation_time=datetime.datetime.strptime(data["creation_time"],
               PreviewTaskAndResults.DATETIME_FORMAT),
          status=PreviewTaskStatus(data["status"]),
          results=data.get("results"),
          id=data["id"],
      )



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import dataclasses
import functools
import operator
from typing import Final

import gaarf

from googleads_housekeeper import views
from googleads_housekeeper.adapters.orm import preview_task_and_results
from googleads_housekeeper.domain.core import (
  exclusion_specification,
  preview_task,
  task,
)
from googleads_housekeeper.domain.external_parsers import external_entity_parser
from googleads_housekeeper.domain.placement_handler import (
  placement_excluder,
  placement_fetcher,
  placement_info_extractor,
)
from googleads_housekeeper.services import enums, unit_of_work

PAGE_SIZE: Final[int] = 40


@dataclasses.dataclass
class RuntimeOptions:
  """Options for configuring, fetching and saving of data."""

  always_fetch_youtube_preview_mode: bool = True
  save_to_db: bool = True


def exclude_placements(
  task_obj: task.Task,
  uow: unit_of_work.AbstractUnitOfWork,
  client,
  runtime_options: RuntimeOptions,
) -> placement_excluder.ExclusionResult:
  """Excludes data from Google Ads based on specified criteria.

  Args:
    task_obj: A Task for identifying placements suitable for exclusion.
    uow: Unit of Work to handle the transaction.
    runtime_options: Options for configuring fetching and saving of data.
  """
  report_fetcher = gaarf.query_executor.AdsReportFetcher(client)
  fetcher = placement_fetcher.PlacementFetcher(report_fetcher)
  excluder = placement_excluder.PlacementExcluder(
    client, task_obj.exclusion_level
  )
  to_be_excluded_placements = find_placements_for_exclusion(
    task_obj, uow, report_fetcher, runtime_options
  )
  if not to_be_excluded_placements:
    return placement_excluder.ExclusionResult()
  customer_ids = to_be_excluded_placements['customer_id'].to_list(
    flatten=True, distinct=True
  )
  placement_exclusion_lists = fetcher.get_placement_exclusion_lists(
    customer_ids
  )
  return excluder.exclude_placements(
    to_be_excluded_placements, placement_exclusion_lists
  )


def _is_already_excluded_placement(
  placement_info: gaarf.report.GaarfRow,
  already_excluded_placements: dict[int, list[int]],
  exclusion_level: enums.ExclusionLevelEnum,
) -> bool:
  """Verifies that a placement has already been added to exclusions.

  Args:
    placement_info:
      GaarfRow containing information on a single placement.
    already_excluded_placements:
      Mapping for placements added as negative criteria on a
      specific level (ACCOUNT, CAMPAIGN, AD_GROUP).
    exclusion_level:
      Level of exclusion (ACCOUNT, CAMPAIGN, AD_GROUP).

  Returns:
    Description of return.

  Raises:
    ValueError: When incorrect exclusion level is specified.
  """
  if exclusion_level == enums.ExclusionLevelEnum.ACCOUNT:
    level_id = placement_info.customer_id
  elif exclusion_level == enums.ExclusionLevelEnum.CAMPAIGN:
    level_id = placement_info.campaign_id
  elif exclusion_level == enums.ExclusionLevelEnum.AD_GROUP:
    level_id = placement_info.ad_group_id
  else:
    raise ValueError(f'Incorrect exclusion_level: {exclusion_level}')
  return placement_info.placement in already_excluded_placements.get(
    level_id, []
  )


def find_placements_for_exclusion(
  task_obj: task.Task,
  uow: unit_of_work.AbstractUnitOfWork,
  report_fetcher: gaarf.query_executor.AdsReportFetcher | None = None,
  runtime_options: RuntimeOptions = RuntimeOptions(),
) -> gaarf.report.GaarfReport | None:
  """Identifies the placements suitable for exclusion.

  Args:
    task_obj: A Task for identifying placements suitable for exclusion.
    uow: Unit of Work to handle the transaction.
    report_fetcher: An instance of AdsReportFetcher to handle API requests.
    runtime_options: Options for configuring fetching and saving of data.

  Returns:
    Report containing placements suitable for exclusion.
  """
  fetcher = placement_fetcher.PlacementFetcher(report_fetcher)
  extractor = placement_info_extractor.PlacementInfoExtractor(uow)
  external_parser = external_entity_parser.ExternalEntitiesParser(uow)
  specification = (
    exclusion_specification.ExclusionSpecification.from_expression(
      task_obj.exclusion_rule
    )
  )
  reports: list[gaarf.report.GaarfReport] = []
  for account in task_obj.accounts:
    account_allowlisted_placements = views.allowlisted_placements_as_tuples(
      uow, account
    )
    if not (
      placements := fetcher.get_placements_for_account(account, task_obj)
    ):
      continue
    already_excluded_placements = fetcher.get_already_excluded_placements(
      account, task_obj.exclusion_level
    )
    if ads_specs := specification.ads_specs_entries:
      placements = ads_specs.apply_specifications(
        placements, include_reason=False, include_matching_placement=False
      )
      if not placements:
        continue
    if runtime_options.always_fetch_youtube_preview_mode:
      youtube_channel_specification = (
        exclusion_specification.ExclusionSpecification(
          specifications=[
            [
              exclusion_specification.YouTubeChannelExclusionSpecificationEntry(
                'subscriberCount > 0'
              )
            ],
            [
              exclusion_specification.YouTubeVideoExclusionSpecificationEntry(
                'likesCount > 0'
              )
            ],
          ]
        )
      )
      external_parser.parse_specification_chain(
        placements, youtube_channel_specification
      )
    if non_ads_specs := specification.non_ads_specs_entries:
      external_parser.parse_specification_chain(placements, non_ads_specs)
    placements = inject_extra_data(
      placements,
      extractor,
      account_allowlisted_placements,
      already_excluded_placements,
      task_obj.exclusion_level,
    )
    if placements := specification.apply_specifications(placements):
      reports.append(placements)
  if reports:
    return functools.reduce(operator.add, reports)
  return None


def inject_extra_data(
  placements: gaarf.report.GaarfReport,
  extractor: placement_info_extractor.PlacementInfoExtractor,
  allowlisted_placements: list[tuple[str, str, str]],
  already_excluded_placements: dict[int, list[int]],
  exclusion_level: enums.ExclusionLevelEnum,
) -> gaarf.report.GaarfReport:
  """Add extra placement information to the placement report.

  Args:
    placements:
      Report containing placement data.
    extractor:
      Responsible for getting parsed placement metadata data from DB.
    allowlisted_placements:
      Info on placements that shouldn't be excluded form a given account.
    already_excluded_placements:
      Mapping between entity_id from specified exclusion level to all
      negative criteria ids.
    exclusion_level: Level of exclusion (ACCOUNT, CAMPAIGN, AD_GROUP).

  Returns:
    Placement data report with new columns
    ('allowlisted', 'already_excluded', 'extra_info').
  """
  has_allowlisted_placements = allowlisted_placements or False
  for placement in placements:
    if has_allowlisted_placements:
      placement['allowlisted'] = (
        placement.name,
        placement.placement_type,
        placement.customer_id,
      ) in allowlisted_placements
    else:
      placement['allowlisted'] = False
    placement['excluded_already'] = _is_already_excluded_placement(
      placement, already_excluded_placements, exclusion_level
    )
    placement['extra_info'] = extractor.extract_placement_info(placement)
  return placements



import datetime
from dataclasses import is_dataclass, asdict
from enum import Enum


def as_serializable_dict(obj):
    """Convert a dataclass to a JSON-serializable dictionary."""
    if not is_dataclass(obj):
        raise TypeError(f"Object {obj} is not a dataclass instance.")

    def serialize(value):
        if isinstance(value, datetime.datetime):
            return value.isoformat()  # Convert datetime to ISO 8601 string
        elif isinstance(value, Enum):
            return value.value  # Convert Enum to its value
        elif isinstance(value, list):
            return [serialize(v) for v in value]  # Recursively serialize lists
        elif isinstance(value, dict):
            return {k: serialize(v) for k, v in
                    value.items()}  # Serialize dictionaries
        elif is_dataclass(value):
            return as_serializable_dict(
                value)  # Recursively serialize nested dataclasses
        return value  # Return other types as-is

    return {key: serialize(value) for key, value in asdict(obj).items()}


# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=C0330, g-bad-import-order
from __future__ import annotations

import abc
from dataclasses import asdict
from typing import Any, Generic, TypeVar

from google.cloud.firestore_v1.base_query import FieldFilter

from googleads_housekeeper.domain.core import task

T = TypeVar('T')


class AbstractRepository(abc.ABC, Generic[T]):
  def __init__(self):
    self.seen: set[T] = set()

  def add(self, entity: T) -> None:
    """Adds an entity to the repository."""
    self._add(entity)

  def delete(self, entity_id: str) -> None:
    """Deletes an entity by its ID."""
    self._delete(entity_id)

  def get(self, entity_id: str) -> T:
    """Retrieves an entity by its ID."""
    return self._get(entity_id)

  def get_by_conditions(self, conditions: dict[str, Any]) -> T:
    """Retrieves an entity based on specific conditions."""
    return self._get_by_conditions(conditions)

  def get_by_condition(self, condition_name: str, condition_value: str) -> T:
    """Retrieves an entity based on specific conditions."""
    return self._get_by_conditions({condition_name: condition_value})

  def list(self) -> list[T]:
    """Retrieves all entities as a list."""
    return self._list()

  def update(self, entity_id: str, update_dict: dict[str, str]) -> T:
    return self._update(entity_id, update_dict)

  @abc.abstractmethod
  def _add(self, entity: T): ...

  @abc.abstractmethod
  def _get(self, entity_id) -> T: ...

  @abc.abstractmethod
  def _get_by_conditions(self, conditions: dict[str, Any]) -> T: ...

  @abc.abstractmethod
  def _list(self) -> list[T]: ...

  @abc.abstractmethod
  def list_with_creation_time_paginated(
    self, limit=20, offset=0
  ) -> (list)[T]: ...

  @abc.abstractmethod
  def _update(self, entity_id, update_dict) -> None: ...

  @abc.abstractmethod
  def _delete(self, entity_id) -> None: ...


class SqlAlchemyRepository(AbstractRepository[T]):
  def __init__(self, session, entity=task.Task):
    super().__init__()
    self.session = session
    self.entity = entity

  def _add(self, entity) -> None:
    self.session.add(entity)

  def _get(self, entity_id) -> T:
    return self.session.query(self.entity).filter_by(id=entity_id).first()

  def _get_by_conditions(self, conditions: dict[str, Any]) -> list[T]:
    query = self.session.query(self.entity)
    for condition_name, condition_value in conditions.items():
      query = query.filter(
        getattr(self.entity, condition_name) == condition_value
      )
    return query.all()

  def _list(self) -> list[T]:
    return self.session.query(self.entity).all()

  def list_with_creation_time_paginated(self, limit=20, offset=0) -> list[T]:
    """Returns a paginated list of entities, ordered by creation time.

    Args:
      limit: The maximum number of entities to retrieve.
      offset: The number of entities to skip before starting retrieval.
    """
    return (
      self.session.query(self.entity)
      .order_by(self.entity.creation_time.desc())
      .offset(offset)
      .limit(limit)
      .all()
    )

  def _update(self, entity_id, update_dict) -> T:
    return (
      self.session.query(self.entity)
      .filter_by(id=entity_id)
      .update(update_dict)
    )

  def _delete(self, entity_id) -> T:
    return self.session.query(self.entity).filter_by(id=entity_id).delete()


class FirestoreRepository(AbstractRepository[T]):
  def __init__(self, client, entity=T):
    super().__init__()
    self.client = client
    self.entity = entity
    self.collection_name = entity.__name__

  def _add(self, entity) -> None:
    if hasattr(entity, 'id'):
      element_id = entity.id
    else:
      element_id = entity._id
    element_dict = {}
    for key, value in asdict(entity).items():
      if hasattr(value, 'name'):
        value = value.name
      element_dict[key] = value
    self.client.collection(self.collection_name).document(str(element_id)).set(
      element_dict
    )

  def _get(self, entity_id: str) -> T:
    doc = self.client.collection(self.collection_name).document(entity_id).get()
    if doc.exists:
      return self.entity(**doc.to_dict())
    return None

  def _get_by_conditions(self, conditions: dict[str, Any]) -> list[T]:
    try:
      query = self.client.collection(self.collection_name)
      for condition_name, condition_value in conditions.items():
        query = query.where(
          filter=FieldFilter(condition_name, '==', condition_value)
        )
      results = [self.entity(**result.to_dict()) for result in query.stream()]
      return results
    except Exception:
      return []

  def _list(self):
    results = self.client.collection(self.collection_name).stream()
    entities = [self.entity(**result.to_dict()) for result in results]
    return entities

  def list_with_creation_time_paginated(self, limit=20, offset=0) -> (list)[T]:
    """Returns a paginated list of entities, ordered by creation time.

    Args:
      limit: The maximum number of entities to retrieve.
       offset: The number of entities to skip before starting retrieval.
    """
    results = (
      self.client.collection(self.collection_name)
      .order_by('creation_time', direction='DESCENDING')
      .offset(offset)
    )
    if limit is not None:
      results = results.limit(limit)
    entities = [self.entity(**result.to_dict()) for result in results.stream()]
    return entities

  def _update(self, entity_id: str, update_dict: dict[str, Any]) -> None:
    if doc := self.client.collection(self.collection_name).document(entity_id):
      doc.update(update_dict)

  def _delete(self, entity_id: str) -> None:
    self.client.collection(self.collection_name).document(entity_id).delete()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for translating exclusion rule into specifications."""

from __future__ import annotations

import dataclasses
import re
from collections.abc import Sequence


@dataclasses.dataclass(frozen=True)
class Rule:
  """Holds information on exclusion rule."""

  exclusion_type: str
  exclusion_rule: str


def generate_rules(raw_rules: str | Sequence[str]) -> list[list[Rule]] | None:
  """Convert raw string rules into a list of Rules.

  Rule as raw string can be expressed in an explicit
  (`exclusion_type:exclusion_rule`), implicit (`exclusion_rule`) format.

  Rules in each of these formats can be represented as sequence or a single
  string. And-rules are separated by `AND` keyword or comma (`,`),
  Or-rules are separated by `OR` keyword.

  If `exclusion_type` is not specified for a given rule it's treated
  as GOOGLE_ADS_INFO.
  """
  if not raw_rules:
    return None
  rules: list[list[Rule]] = []
  default_exclusion_type = 'GOOGLE_ADS_INFO'
  if isinstance(raw_rules, str):
    raw_rules = _format_raw_rules(raw_rules)
  for rule in raw_rules:
    current_exclusion_type = None
    types = rule.split(',') if ',' in rule else rule.split(' AND ')
    rule_entry: list[Rule] = []
    for type_ in types:
      spec_ = type_.split(':')
      if len(spec_) == 1:
        if not current_exclusion_type:
          exclusion_type = default_exclusion_type
        else:
          exclusion_type = str(current_exclusion_type)
        expression_position = 0
      else:
        exclusion_type = spec_[0]
        expression_position = 1
      current_exclusion_type = str(exclusion_type)
      exclusion_rule = spec_[expression_position].rstrip()
      rule_entry.append(
        Rule(
          exclusion_type.strip(),
          _convert_exclusion_rule(exclusion_rule.strip()),
        )
      )
    rules.append(rule_entry)
  return rules


def _format_raw_rules(raw_rules: str) -> list[str]:
  """Applies formatting to raw string rule."""
  # Removes brackets.
  raw_rules = re.sub('[(|)]', '', raw_rules)
  # Adds whitespace in front of operators.
  raw_rules = re.sub('([>|<|>=|<=|!=|=])', r' \1 ', raw_rules)
  # Trims extra whitespace.
  raw_rules = re.sub(' +', ' ', raw_rules)
  # Split rules by OR operator.
  return raw_rules.split(' OR ')


def _convert_exclusion_rule(exclusion_rule: str) -> str:
  """Handles conversion of composite rule into atomic one."""
  if 'letter_set' in exclusion_rule:
    return exclusion_rule.replace(
      'letter_set latin_only',
      "regexp '^[a-zA-Z0-9\s\W]*$'",
    ).replace('letter_set no_latin', "regexp '^[^a-zA-Z]*$'")
  return exclusion_rule



from __future__ import annotations

# pylint: disable=C0330
import pathlib
from itertools import chain

from setuptools import find_packages, setup

HERE = pathlib.Path(__file__).parent

README = (HERE / 'README.md').read_text()

EXTRAS_REQUIRE = {
  'gcp': [
    'google-cloud-firestore',
    'google-cloud-pubsub',
    'appengine-python-standard>=1.0.0',
  ]
}
EXTRAS_REQUIRE['full'] = list(set(chain(*EXTRAS_REQUIRE.values())))

setup(
  name='googleads-housekeeper',
  version='0.1.0.dev.86',
  long_description=README,
  long_description_content_type='text/markdown',
  author='Google Inc. (gTech gPS CSE team)',
  author_email='no-reply@google.com',
  license='Apache 2.0',
  python_requires='>3.8',
  packages=find_packages(),
  install_requires=[
    'google-ads-api-report-fetcher[pandas]>=1.15.3',
    'SQLAlchemy==1.4.46',
    'beautifulsoup4',
    'google-api-python-client',
    'psycopg2-binary',
    'croniter',
    'aiohttp',
    'slack_sdk',
    'numpy',
  ],
  extras_require=EXTRAS_REQUIRE,
  setup_requires=['pytest-runner'],
  tests_requires=['pytest', 'pytest-mock'],
)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for converting legacy task v1 to v2."""

from __future__ import annotations

from datetime import datetime

from googleads_housekeeper.domain.core import task
from googleads_housekeeper.services import enums


class TaskAdapter:
  """Maps a V1 Task (dict) into a V2 Task object.

  Attributes:
      task_v1_data: A dictionary containing data of a V1 Task object.
  """

  def __init__(self, task_v1_data: dict) -> None:
    """Initializes a TaskAdapter object.

    Args:
        task_v1_data:
            A dictionary containing data of a V1 Task object.
    """
    self.task_v1_data = task_v1_data

  def from_task_v1(self) -> task.Task:
    """Converts a V1 Task into a V2 Task object.

    Returns:
        Task: A V2 Task object.
    """
    task_dict = {
      'name': self.task_v1_data['task_name'],
      'exclusion_rule': self._build_condition_rules(),
      'customer_ids': self.task_v1_data['customer_id'],
      'date_range': int(self.task_v1_data.get('lookback_days', 0)),
      'from_days_ago': int(self.task_v1_data.get('from_days_ago', 0)),
      'exclusion_level': enums.ExclusionLevelEnum.AD_GROUP,
      'placement_types': self._build_placement_types(),
      'output': self._build_task_output(),
      'status': task.TaskStatus.ACTIVE,
      'schedule': self.task_v1_data.get('schedule', 0),
      'creation_date': datetime.strptime(
        self.task_v1_data['date_created'], '%Y-%m-%d'
      ),
    }
    return task.Task(**task_dict)

  def _build_placement_types(self) -> str:
    placement_types = set()
    if self.task_v1_data.get('gads_data_youtube'):
      placement_types.add('YOUTUBE_VIDEO')
      placement_types.add('YOUTUBE_CHANNEL')
    if self.task_v1_data.get('gads_data_display'):
      placement_types.add('WEBSITE')
    return ', '.join(sorted(placement_types))

  def _build_condition_rules(self) -> str:
    youtube_filters = self._build_youtube_filters()
    google_ads_filters = self._build_gads_filters()
    return ' AND '.join([google_ads_filters] + youtube_filters)

  def _build_youtube_filters(self) -> list[str]:
    youtube_filters: list[str] = []
    yt_country_operator = self.task_v1_data.get('yt_country_operator')
    yt_country_value = self.task_v1_data.get('yt_country_value')
    yt_language_operator = self.task_v1_data.get('yt_language_operator')
    yt_language_value = self.task_v1_data.get('yt_language_value')
    yt_std_character = self.task_v1_data.get('yt_std_character')
    yt_subscriber_operator = self.task_v1_data.get('yt_subscriber_operator')
    yt_subscriber_value = self.task_v1_data.get('yt_subscriber_value')
    yt_video_operator = self.task_v1_data.get('yt_video_operator')
    yt_video_value = self.task_v1_data.get('yt_video_value')
    yt_view_operator = self.task_v1_data.get('yt_view_operator')
    yt_view_value = self.task_v1_data.get('yt_view_value')

    if yt_subscriber_operator is not None and yt_subscriber_value is not None:
      youtube_filters.append(
        'YOUTUBE_CHANNEL_INFO:subscriberCount '
        f'{yt_subscriber_operator}, {yt_subscriber_value},'
      )
    if yt_view_operator is not None and yt_view_value is not None:
      youtube_filters.append(
        'YOUTUBE_CHANNEL_INFO:yt_view_operator '
        f'{yt_view_operator} {yt_view_value}'
      )
    if yt_video_operator is not None and yt_video_value is not None:
      youtube_filters.append(
        'YOUTUBE_CHANNEL_INFO:yt_video_operator '
        f'{yt_video_operator} {yt_video_value}'
      )
    if yt_country_operator is not None and yt_country_value is not None:
      youtube_filters.append(
        'YOUTUBE_CHANNEL_INFO:country '
        f'{yt_country_operator} {yt_country_value}'
      )
    if yt_language_operator is not None and yt_language_value is not None:
      youtube_filters.append(
        'YOUTUBE_CHANNEL_INFO:defaultLanguage '
        f'{yt_language_operator}, {yt_language_value},'
      )
    if yt_std_character is not None and yt_std_character != '':
      youtube_filters.append(
        'YOUTUBE_CHANNEL_INFO:title letter_set latin_only'
        if yt_std_character == '1'
        else 'YOUTUBE_CHANNEL_INFO:title letter_set no_latin'
        if yt_std_character == '0'
        else 'YOUTUBE_CHANNEL_INFO:title letter_set unknown'
      )
    return youtube_filters

  def _build_gads_filters(self) -> str:
    if google_ads_filters := self.task_v1_data.get('gads_filter'):
      return google_ads_filters.replace('metrics.', 'GOOGLE_ADS_INFO:')
    return ''

  def _build_task_output(self) -> task.TaskOutput:
    schedule = int(self.task_v1_data.get('schedule', 0))
    email_alerts = bool(self.task_v1_data.get('email_alerts', False))
    if email_alerts and schedule == 1:
      return task.TaskOutput.EXCLUDE_AND_NOTIFY
    if not email_alerts and schedule == 1:
      return task.TaskOutput.EXCLUDE
    if email_alerts and schedule == 0:
      return task.TaskOutput.NOTIFY
    return task.TaskOutput.EXCLUDE_AND_NOTIFY
